#include "ft_nm.h"

void help(char *bin_name) {
	ft_printf("sss", "Usage: ", bin_name, " [option(s)] [file(s)]\n");
	ft_putstr(
			" List symbols in [file(s)] (a.out by default).\n"
			" The options are:\n"
			"  -a Display debugger-only symbols\n"
			"  -g Display only external symbols\n"
			"  -p Do not sort the symbols\n"
			"  -r Reverse the sense of the sort\n"
			"  -u Display only undefined symbols\n"
			"  -h Display this information\n");
	ft_printf("ss", bin_name, ": supported targets: elf64-x86-64 elf32-x86-64\n");
	ft_putstr("Report bugs to gaetanbrochard@protonmail.com\n");
}

int parse_options(char *options, t_env *env, bool *use_options) {
	int len = ft_strlen(options);
	if (len == 2 && options[1] == '-') {
		*use_options = false;
		return 0;
	}
	for (int i = 1; i < len; i++) {
		switch (options[i]) {
			case 'h':
				help(env->bin_name);
				return -2;
				break;
			case 'a':
				if (env->verbosity < ALL || env->verbosity == DEFAULT_VERBOSITY)
					env->verbosity = ALL;
				break;
			case 'g':
				if (env->verbosity < GLOBAL)
					env->verbosity = GLOBAL;
				break;
			case 'u':
				if (env->verbosity < UNDEFINED)
					env->verbosity = UNDEFINED;
				break;
			case 'r':
				if (env->ordering < REVERSE)
					env->ordering = REVERSE;
				break;
			case 'p':
				if (env->ordering < NOSORT)
					env->ordering = NOSORT;
				break;
			default:
				return -1;
		}
	}
	return 0;
}

int parse_arguments(int ac, char **av, t_env *env) {
	bool use_options = true;
	if (ac > 0) {
		env->bin_name = ft_strdup("nm");
	}
	for (int i = 1; i < ac; i++) {
		if (ft_strlen(av[i]) > 0 && av[i][0] == '-' && use_options) {
			int ret = parse_options(av[i], env, &use_options);
			if (ret != 0)
				return ret;
		} else {
			if (env->list_len == 0) {
				env->list = ft_lstnew(av[i], ft_strlen(av[i]) + 1);
			} else {
				ft_lstadd_back(&env->list, ft_lstnew(av[i], ft_strlen(av[i]) + 1));
			}
			env->list_len++;
		}
	}
	return 0;
}

int nm(char *path, t_verbosity verbosity, t_ordering ordering) {
	int fd = open(path, O_RDONLY);
	if (fd == -1) {
		ft_nm_error(path);
	}

	struct stat stat;
	ft_bzero(&stat, sizeof(stat));
	if (fstat(fd, &stat) == -1) {
		ft_nm_error(path);
	}

	t_mapped_file mapped_file;

	mapped_file.ptr = mmap(NULL, stat.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	mapped_file.size = stat.st_size;
	if (mapped_file.ptr != MAP_FAILED) {
		void *ident_ptr = fetch(mapped_file, 0, EI_NIDENT);
		unsigned char ident[EI_NIDENT];
		if (!ident_ptr) {
			ft_printf("sss", "nm: ", path, ": file format not recognized\n");
			munmap(mapped_file.ptr, mapped_file.size);
			return 0;

		}
		ft_memcpy(ident, ident_ptr, EI_NIDENT);
		if (check_ident(ident) == FT_NM_FAILURE) {
			ft_printf("sss", "nm: ", path, ": file format not recognized\n");
			munmap(mapped_file.ptr, mapped_file.size);
			return 0;
		}
		if (ident[EI_CLASS] == ELFCLASS32) {
			if (nm32(mapped_file, path, verbosity, ordering) == FT_NM_FAILURE) {
				ft_printf("sss", "nm: ", path, ": file format not recognized\n");
				munmap(mapped_file.ptr, mapped_file.size);
				return 0;
			}
		}
		else if (ident[EI_CLASS] == ELFCLASS64) {
			if (nm64(mapped_file, path, verbosity, ordering) == FT_NM_FAILURE) {
				ft_printf("sss", "nm: ", path, ": file format not recognized\n");
				munmap(mapped_file.ptr, mapped_file.size);
				return 0;
			}
		}
	}
	return 0;
}

void lst_free_func(void *content, size_t content_size) {
	(void)content_size;
	free(content);
}

void free_env(t_env env) {
	ft_lstdel(&(env.list), &lst_free_func);
	if (env.bin_name)
		free(env.bin_name);
}

int main(int ac, char **av) {
	t_env env;
	env.verbosity = DEFAULT_VERBOSITY;
	env.ordering = DEFAULT_ORDERING;
	env.list_len = 0;
	env.list = NULL;
	int error = parse_arguments(ac, av, &env);
	if (env.list_len == 0) {
		env.list = ft_lstnew("a.out", 6);
		env.list_len = 1;
	}
	if (error) {
		free_env(env);
		return error;
	}
	if (env.list_len == 1) {
		nm(env.list->content, env.verbosity, env.ordering);
		free_env(env);
		return 0;
	}
	t_list *tmp = env.list;
	while (tmp) {
		ft_printf("\ns:\n", env.list->content);
		nm(env.list->content, env.verbosity, env.ordering);
		tmp = tmp->next;
	}
	free_env(env);
	return 0;
}
